@@ -16,7 +16,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: silcauth.c,v 1.28 2002/11/22 20:06:40 priikone Exp $ */
+/* $Id: silcauth.c,v 1.28.2.1 2002/12/14 13:18:32 priikone Exp $ */
 
 #include "silcincludes.h"
 #include "silcauth.h"
@@ -241,7 +241,7 @@ SilcBuffer silc_auth_public_key_auth_generate(SilcPublicKey public_key,
 					      const void *id, SilcIdType type)
 {
   unsigned char *randomdata;
-  unsigned char auth_data[2048];
+  unsigned char auth_data[2048 + 1];
   SilcUInt32 auth_len;
   unsigned char *tmp;
   SilcUInt32 tmp_len;
